#pragma rtGlobals=1		// Use modern global access method.

// This procedure file loads "Universal File" version 58 formatted files.
// This format was created by Structural Dynamics Research Corporation in the 1960's and 1970's
// and has been extended many times. It is used by the experimental dynamics (vibration and acoustic) community.
// A general discussion of the format can be found here:
//	http://www.sdrl.uc.edu/universal-file-formats-for-modal-analysis-testing-1

// The version 58 "Unversal file" format is described here:
//	http://www.sdrl.uc.edu/universal-file-formats-for-modal-analysis-testing-1/file-format-storehouse/unv_0058.htm

// This code does not load version 58b which is a mix of ASCII and binary or any format other than version 58.

// The version 58 file consists of a variable number of sections. Each section consists of:
//	A two-line preamble that identifies the file format
//	An 11-line header
//	A variable number of lines of data in ASCII text format
//
// Each section represents the data for one dataset. A dataset may consist of:
//	An evenly-spaced array. This is analogous to an Igor waveform.
//	An unevenly-spaced array. This is analogous to an XY pair in Igor.
// Each section may contain real or complex data.
// Each section may contain single-precision or double-precision data.
// Combining these variations gives 8 different section types. These types are
// described under "Record 12" in unv_0058.htm.
//
// The data in a section is organized into 3, 4 or 6 columns depending on the
// type of the data represented by the section. The data is row-oriented.
// For sections that represent more than one array, for example XY or complex sections,
// the data is interleaved.

// The LoadUniversalVersion58File function loads data from a version 58 universal file.
// It creates header a wave and one or more data waves for each section in the file.

// The LoadUniversalVersion58FilePI function is a programmatic interface to the file loader
// for use by programmers who want to customize it to provide custom wave names, for example.

// We will first describe the default behavior provided by LoadUniversalVersion58File.
// Then we will describe the customizations possible using LoadUniversalVersion58FilePI.

// The LoadUniversalVersion58File function first creates a new data folder using a name based on the file name.
// The waves are loaded in the new data folder.

// The header wave for a given section is a text wave. It is named "uni<n>Header"
// where <n> is the zero-based section number.
// The header wave is used by the package during the loading process. You may not
// need the header waves. However, you may want to examine a header wave in
// a table or use it programmatically (discussed later).

// The data for each section is loaded into one or more waves as follows:
//
//	Real, single-precision or double-precision, evenly-spaced section (tested)
//		Loaded into a wave named "uni<n>" where <n> represents the section number.
//
//	Real, single-precision or double-precision, unevenly-spaced section (untested)
//		Loaded into two waves named "uni<n>X" and "uni<n>Y" where <n> represents the section number.
//
//	Complex, single-precision or double-precision, evenly-spaced section (tested)
//		If the CreateComplex parameter is 0
//			Loaded into two real waves named "uni<n>Real" and "uni<n>Imag" where <n> represents the section number.
//		If the CreateComplex parameter is 1
//			Loaded into one complex wave named "uni<n>Complex" where <n> represents the section number.
//
//	Complex, single-precision or double-precision, unevenly-spaced section (untested)
//		If the CreateComplex parameter is 0
//			Loaded into three real waves named "uni<n>X", "uni<n>Real" and "uni<n>Imag" where <n> represents the section number.
//		If the CreateComplex parameter is 1
//			Loaded into one real wave named "uni<n>X" and one complex wave named "uni<n>Complex" where <n> represents the section number.
//
// The types marked "untested" are untested because we don't have a sample file containing that type of data.
//

// Parameters to LoadUniversalVersion58File function
//
//	pathName: Name of symbolic path or "".
//
//	filePath: File name, partial path or full path to the file.
//
//	CreateComplex:
//		If CreateComplex is 0, when a complex section is loaded, two real waves are created named uni<n>Real and uni<n>Imag.
//		If CreateComplex is 1, when a complex section is loaded, one complex wave is created named uni<n>Complex
//
//	Quiet: If 0, a message is printed to history for section loaded.
//	If 1, no message is printed but error messages are still printed.
//
//	ReuseDataFolder: The package creates a data folder in which the output waves are created.
//	The data folder name is based on the name of the file being loaded.
//	"Reusing" means creating the waves in the data folder, possibly overwriting waves
//	already in the data folder. Also resuing could leave you with a data folder that
//	contains a mix of waves from different files.
//	If the data folder already exist:
//		If ReuseDataFolder is 0: The package asks whether to reuse it.
//		If ReuseDataFolder is 1: The package automatically reuses it without asking.

// Programmatic Interface
//
// If you are using the package programmatically, you can control the name of the output data
// folder and the names of the created waves by using LoadUniversalVersion58FilePI instead of
// LoadUniversalVersion58File. The LoadUniversalVersion58File function itself calls
// LoadUniversalVersion58FilePI and therefore provides an example of how to do this.
//
// Start by copying the LoadUniversalVersion58File at the bottom of this file to your own procedure file.
// Make changes in your own procedure file so you can update this one without losing your customizations.
//
// You can control the name of the output data folder by providing your own GenerateDFName function.
// You can create your function by copying the GenerateUniversal58DFName function, renaming it, and customizing it. 
//
// You can control the names of the output waves by providing your own GenerateWaveName function.
// You can create your function by copying the GenerateUniversal58WaveName function, renaming it, and customizing it. 
//
// When using the file loader programmatically, you may find it useful to load the header wave
// into a UniversalFile58HeaderInfo structure by calling GetUniversal58HeaderInfo.
//
// The LoadUniversalVersion58FilePI function also gives you access, through the dfr field, to the
// data folder containing the output waves. You can use this for post-processing the waves.
//
// If you are using these procedures programmatically, you may want your own menu item
// in the Data->Load Waves submenu. In that case, copy the menu definition below to your
// own file and comment the one below out.

Menu "Load Waves"
	"Load Universal Version 58 File...", LoadUniversalVersion58File("", "", 0, 1, 0)
End

static Function/S RemoveWhitespace(input)
	String input
	
	String output = ReplaceString(" ", input, "")	// Remove spaces
	output = ReplaceString("\t", output, "")		// Remove tabs
	return output
End

Structure UniversalFile58HeaderInfo
	String versionString				// e.g., "58"
	
	String generalDescription		// Record 1
	String record2					// Record 2 - not documented
	String dateTimeStr				// Record 3
	String record4					// Record 4 - not documented
	String record5					// Record 5 - not documented

	// Start of fields from record 6 - DOF Identification
	Variable functionType
	Variable functionIDNumber
	Variable versionOrSequenceNumber
	Variable loadCaseIDNumber			// 0=Single Point Excitation
	String responseEntity
	Variable responseNode
	Variable responseDirection
	String referenceEntity
	Variable referenceNode
	Variable referenceDirection
	// End of fields from record 6

	// Start of fields from record 7
	Variable ordinateDataType			// 2=real SP, 4=real DP, 5=complex SP, 6=complex DP
	Variable numDataPairsOrValues		// Num XY pairs for uneven abcissa, num values for even abcissa
	Variable abcissaSpacing				// 0=uneven, 1=even (no abcissa values stored)
	Variable abcissaMin					// x0 - 0.0 if uneven abcissa
	Variable abcissaIncrement			// dx - 0.0 if uneven abcissa
	Variable zAxisValue					// 0.0 if unused
	// End of fields from record 7

	// Start of fields from record 8 - Abcissa data characteristics
	Variable abcissaDataCharacteristics
	Variable abcissaLengthUnitsExponent
	Variable abcissaForceUnitsExponent
	Variable abcissaTempUnitsExponent
	String abcissaAxisLabel				// NONE if not used
	String abcissaAxisUnitsLabel			// NONE if not used
	// End of fields from record 8

	// Start of fields from record 9 - Ordinate data characteristics
	Variable ordinateDataCharacteristics
	Variable ordinateLengthUnitsExponent
	Variable ordinateForceUnitsExponent
	Variable ordinateTempUnitsExponent
	String ordinateAxisLabel				// NONE if not used
	String ordinateAxisUnitsLabel		// NONE if not used
	// End of fields from record 9

	// Start of fields from record 10 - Ordinate denominator data characteristics
	Variable ordDenomDataCharacteristics
	Variable ordDenomLengthUnitsExponent
	Variable ordDenomForceUnitsExponent
	Variable ordDenomTempUnitsExponent
	String ordDenomAxisLabel			// NONE if not used
	String ordDenomAxisUnitsLabel		// NONE if not used
	// End of fields from record 10

	// Start of fields from record 11 - Z-axis data characteristics
	Variable zAxisDataCharacteristics
	Variable zAxisLengthUnitsExponent
	Variable zAxisForceUnitsExponent
	Variable zAxisTempUnitsExponent
	String zAxisAxisLabel				// NONE if not used
	String zAxisAxisUnitsLabel			// NONE if not used
	// End of fields from record 11
EndStructure

static Function GetUniversal58HeaderInfo(sectionNumber, startLine, headerWave, header)
	Variable sectionNumber	// Zero-based section number.
	Variable startLine		// Zero-based line on which the section starts.
	Wave/T headerWave							// Contains 13 lines of header info
	STRUCT UniversalFile58HeaderInfo &header		// Parsed header info returned here
	
	// Parse preamble lines and verify that this is the right kind of file
	
	// The file starts with two preamble lines that are not mentioned on the web page listed above
	Variable numPreambleLines = 2

	String firstPreambleLine = headerWave[0]
	firstPreambleLine = RemoveWhitespace(firstPreambleLine)
	if (CmpStr(firstPreambleLine,"-1") != 0)
		Printf "First preamble line of universal file, section %d, line %d, must contain \"-1\"\r", sectionNumber, startLine
		return -1
	endif
	
	String secondPreambleLine = headerWave[1]
	secondPreambleLine = RemoveWhitespace(secondPreambleLine)
	if (CmpStr(secondPreambleLine,"58") != 0)
		Printf "This is not a version 58 universal file, section %d, line %d\r", sectionNumber, startLine
		return -1
	endif
	header.versionString = secondPreambleLine
	
	// Parse header lines.
	// The "header lines", properly speaking, follow the two preamble lines.
	// We use the terminology "line1", "line2" following the 1-based
	// convention of the web page listed above
	
	String line1 = headerWave[2]
	header.generalDescription = line1
	
	String line2 = headerWave[3]		// The meaning of Record 2 is not documented
	header.record2 = line2

	String line3 = headerWave[4]		// Generally date/time
	header.dateTimeStr = line3
	
	String line4 = headerWave[5]		// The meaning of Record 4 is not documented
	header.record4 = line4
	
	String line5 = headerWave[6]		// The meaning of Record 5 is not documented
	header.record5 = line5
	
	String line6 = headerWave[7]		// Record 6 - 10 fields
	Variable functionType, functionIDNumber, versionOrSequenceNumber, loadCaseIDNumber
	String responseEntity
	Variable responseNode, responseDirection
	String referenceEntity
	Variable referenceNode, referenceDirection
	sscanf line6, " %d %d %d %d %s %d %d %s %d %d", functionType, functionIDNumber, versionOrSequenceNumber, loadCaseIDNumber, responseEntity, responseNode, responseDirection, referenceEntity, referenceNode, referenceDirection
	header.functionType = functionType
	header.functionIDNumber = functionIDNumber
	header.versionOrSequenceNumber = versionOrSequenceNumber
	header.loadCaseIDNumber = loadCaseIDNumber
	header.responseEntity = responseEntity
	header.responseNode = responseNode
	header.responseDirection = responseDirection
	header.referenceEntity = referenceEntity
	header.referenceNode = referenceNode
	header.referenceDirection = referenceDirection

	String line7 = headerWave[8]		// Record 7 - 6 fields
	Variable ordinateDataType, numDataPairsOrValues
	Variable abcissaSpacing, abcissaMin, abcissaIncrement
	Variable zAxisValue
	sscanf line7, " %d %d %d %g %g %g", ordinateDataType, numDataPairsOrValues, abcissaSpacing, abcissaMin, abcissaIncrement, zAxisValue
	header.ordinateDataType = ordinateDataType
	header.numDataPairsOrValues = numDataPairsOrValues
	header.abcissaSpacing = abcissaSpacing
	header.abcissaMin = abcissaMin
	header.abcissaIncrement = abcissaIncrement
	header.zAxisValue = zAxisValue

	String line8 = headerWave[9]		// Record 8 - 6 fields - abcissa data characteristics
	Variable dataCharacteristics, lengthUnitsExponent, forceUnitsExponent, tempUnitsExponent
	String axisLabel, axisUnitsLabel
	sscanf line8, " %d %d %d %d %s %s", dataCharacteristics, lengthUnitsExponent, forceUnitsExponent, tempUnitsExponent, axisLabel, axisUnitsLabel
	header.abcissaDataCharacteristics = dataCharacteristics
	header.abcissaLengthUnitsExponent = lengthUnitsExponent
	header.abcissaForceUnitsExponent = forceUnitsExponent
	header.abcissaTempUnitsExponent = tempUnitsExponent
	header.abcissaAxisLabel = axisLabel
	header.abcissaAxisUnitsLabel = axisUnitsLabel

	String line9 = headerWave[10]		// Record 9 - 6 fields - ordinate data characteristics
	sscanf line9, " %d %d %d %d %s %s", dataCharacteristics, lengthUnitsExponent, forceUnitsExponent, tempUnitsExponent, axisLabel, axisUnitsLabel
	header.ordinateDataCharacteristics = dataCharacteristics
	header.ordinateLengthUnitsExponent = lengthUnitsExponent
	header.ordinateForceUnitsExponent = forceUnitsExponent
	header.ordinateTempUnitsExponent = tempUnitsExponent
	header.ordinateAxisLabel = axisLabel
	header.ordinateAxisUnitsLabel = axisUnitsLabel

	String line10 = headerWave[11]		// Record 10 - 6 fields - ordinate denominator characteristics
	sscanf line10, " %d %d %d %d %s %s", dataCharacteristics, lengthUnitsExponent, forceUnitsExponent, tempUnitsExponent, axisLabel, axisUnitsLabel
	header.ordDenomDataCharacteristics = dataCharacteristics
	header.ordDenomLengthUnitsExponent = lengthUnitsExponent
	header.ordDenomForceUnitsExponent = forceUnitsExponent
	header.ordDenomTempUnitsExponent = tempUnitsExponent
	header.ordDenomAxisLabel = axisLabel
	header.ordDenomAxisUnitsLabel = axisUnitsLabel

	String line11 = headerWave[12]		// Record 11 - 6 fields - Z-axis data characteristics
	sscanf line11, " %d %d %d %d %s %s", dataCharacteristics, lengthUnitsExponent, forceUnitsExponent, tempUnitsExponent, axisLabel, axisUnitsLabel
	header.zAxisDataCharacteristics = dataCharacteristics
	header.zAxisLengthUnitsExponent = lengthUnitsExponent
	header.zAxisForceUnitsExponent = forceUnitsExponent
	header.zAxisTempUnitsExponent = tempUnitsExponent
	header.zAxisAxisLabel = axisLabel
	header.zAxisAxisUnitsLabel = axisUnitsLabel
	
	return 0
End

// DetermineDataCharacteristics(header, numDataColumns, numDataLines, defaultFieldWidth, column
// Determines characteristics of the section that are needed to load it.
static Function DetermineDataCharacteristics(header, numDataColumns, numDataLines, defaultFieldWidth, columnInfo)
	STRUCT UniversalFile58HeaderInfo &header	// Parsed header info
	Variable& numDataColumns					// Output
	Variable& numDataLines						// Output
	Variable& defaultFieldWidth					// Output
	String& columnInfo							// Output

	columnInfo = ""
	
	Variable numDataPairsOrValues = header.numDataPairsOrValues		// Num XY pairs for uneven abcissa, num values for even abcissa
	Variable totalNumberOfNumbers = numDataPairsOrValues			// Adjusted below

	Variable ordinateDataType = header.ordinateDataType	// 2=real SP, 4=real DP, 5=complex SP, 6=complex DP
	Variable abcissaSpacing = header.abcissaSpacing		// 0=uneven, 1=even (no abcissa values stored)
	
	Variable isComplex = ordinateDataType==5 || ordinateDataType==6
	Variable isReal = !isComplex
	Variable isDouble = ordinateDataType==4 || ordinateDataType==6
	Variable isSingle = !isDouble
	Variable isUneven = abcissaSpacing == 0
	Variable isEven = !isUneven
	
	// This follows the 8 cases laid out for Record 12 in the version 58 documentation. See web page referenced above.

	if (isReal && isSingle && isEven)			// Case 1: 6E13.5
		numDataColumns = 6
		defaultFieldWidth = 13
		columnInfo = "N=uniRaw0,T=2;N=uniRaw1,T=2;N=uniRaw2,T=2;N=uniRaw3,T=2;N=uniRaw4,T=2;N=uniRaw5,T=2;"
	endif

	if (isReal && isSingle && isUnEven)		// Case 2: 6E13.5
		totalNumberOfNumbers *= 2		// Because it is XY
		numDataColumns = 6
		defaultFieldWidth = 13
		columnInfo = "N=uniRaw0,T=2;N=uniRaw1,T=2;N=uniRaw2,T=2;N=uniRaw3,T=2;N=uniRaw4,T=2;N=uniRaw5,T=2;"
	endif
	
	if (isComplex && isSingle && isEven)		// Case 3: 6E13.5
		totalNumberOfNumbers *= 2		// Because it is complex
		numDataColumns = 6
		defaultFieldWidth = 13
		columnInfo = "N=uniRaw0,T=2;N=uniRaw1,T=2;N=uniRaw2,T=2;N=uniRaw3,T=2;N=uniRaw4,T=2;N=uniRaw5,T=2;"
	endif
	
	if (isComplex && isSingle && isUneven)	// Case 4: 6E13.5
		totalNumberOfNumbers *= 3		// Because it is complex and XY
		numDataColumns = 6
		defaultFieldWidth = 13
		columnInfo = "N=uniRaw0,T=2;N=uniRaw1,T=2;N=uniRaw2,T=2;N=uniRaw3,T=2;N=uniRaw4,T=2;N=uniRaw5,T=2;"
	endif

	if (isReal && isDouble && isEven)			// Case 5: 4E20.12
		numDataColumns = 4
		defaultFieldWidth = 20
		columnInfo = "N=uniRaw0,T=4;N=uniRaw1,T=4;N=uniRaw2,T=4;N=uniRaw3,T=4;"
	endif

	if (isReal && isDouble && isUnEven)		// Case 6: 2(E13.5,E20.12)
		totalNumberOfNumbers *= 2		// Because it is XY
		numDataColumns = 4
		defaultFieldWidth = 20
		columnInfo = "N=uniRaw0,W=13,T=4;N=uniRaw1,W=20,T=4;N=uniRaw2,W=13,T=4;N=uniRaw3,W=20,T=4;"
	endif
	
	if (isComplex && isDouble && isEven)	// Case 7: 4E20.12
		totalNumberOfNumbers *= 2		// Because it is complex
		numDataColumns = 4
		defaultFieldWidth = 20
		columnInfo = "N=uniRaw0,T=4;N=uniRaw1,T=4;N=uniRaw2,T=4;N=uniRaw3,T=4;"
	endif
	
	if (isComplex && isDouble && isUneven)	// Case 8: E13.5,2E20.12
		totalNumberOfNumbers *= 3		// Because it is complex and XY
		numDataColumns = 3
		defaultFieldWidth = 20
		columnInfo = "N=uniRaw0,W=13,T=4;N=uniRaw1,W=20,T=4;N=uniRaw2,W=20,T=4;"
	endif

	numDataLines = ceil(totalNumberOfNumbers/numDataColumns)	
	return 0
End

// GenerateUniversal58WaveName(header, sectionNumber, role)
// Returns the name to use for a wave of the type indicated by the input parameters.
// role specifies the kind of wave we are creating and contributes to determining what name to use.
//	0: Real waveform
//	1: X wave of XY pair
//	2: Y wave of XY pair
//	3: Real part of complex pair
//	4: Imaginary part of complex pair
//	5: Complex waveform
//	6: Header wave
Function/S GenerateUniversal58WaveName(header, sectionNumber, role)
	STRUCT UniversalFile58HeaderInfo &header	// Parsed header info
	Variable sectionNumber						// The zero-based section number
	Variable role								// See comments above
	
	String sectionNumberStr = num2istr(sectionNumber)
	
	String name

	switch(role)
		case 0:										// Real waveform
			name = "uni" + sectionNumberStr
			break
		case 1:										//  X wave of XY pair
			name = "uni" + sectionNumberStr + "X"
			break
		case 2:										// Y wave of XY pair
			name = "uni" + sectionNumberStr + "Y"
			break
		case 3:										// Real part of complex pair
			name = "uni" + sectionNumberStr + "Real"
			break
		case 4:										// Imaginary part of complex pair
			name = "uni" + sectionNumberStr + "Imag"
			break
		case 5:										// Complex waveform
			name = "uni" + sectionNumberStr + "Complex"
			break
		case 6:										// Header wave
			name = "uni" + sectionNumberStr + "Header"
			break
		default:
			Abort "Bug in GenerateUniversal58WaveName"
			break
	endswitch
	
	return name
End

static Function CreateOutputWaves(header, rawWaveList, sectionNumber, startLine, createComplex, GenerateWaveName, outputWaveList)
	STRUCT UniversalFile58HeaderInfo &header	// Parsed header info
	String rawWaveList							// Semicolon-separated list of raw loaded waves
	Variable sectionNumber		// Zero-based section number.
	Variable startLine			// Zero-based line on which the section starts.
	Variable createComplex		// 0: Create two real waves for complex data; 1: Create one complex wave for complex data
	FUNCREF GenerateUniversal58WaveName GenerateWaveName	// Function for naming waves patterned after GenerateUniversal58WaveName
	String& outputWaveList		// Output - list of final waves

	Variable numPointsInSection = header.numDataPairsOrValues	// Num XY pairs for uneven abcissa, num values for even abcissa

	Variable ordinateDataType = header.ordinateDataType	// 2=real SP, 4=real DP, 5=complex SP, 6=complex DP
	Variable abcissaSpacing = header.abcissaSpacing		// 0=uneven, 1=even (no abcissa values stored)
	Variable abcissaMin = header.abcissaMin
	Variable abcissaIncrement = header.abcissaIncrement
	
	String abcissaUnits = header.abcissaAxisUnitsLabel
	if (CmpStr(abcissaUnits,"NONE") == 0)
		abcissaUnits = ""
	endif
	if (CmpStr(abcissaUnits,"Sec") == 0)
		abcissaUnits = "s"
	endif
	String ordinateUnits = header.ordinateAxisUnitsLabel
	if (CmpStr(ordinateUnits,"NONE") == 0)
		ordinateUnits = ""
	endif
	if (CmpStr(ordinateUnits,"Sec") == 0)
		ordinateUnits = "s"
	endif
	
	Variable isComplex = ordinateDataType==5 || ordinateDataType==6
	Variable isReal = !isComplex
	Variable isDouble = ordinateDataType==4 || ordinateDataType==6
	Variable isSingle = !isDouble
	Variable isUneven = abcissaSpacing == 0
	Variable isEven = !isUneven
	
	outputWaveList = ""
	
	// This follows the 8 cases laid out for Record 12 in the version 58 documentation. See web page referenced above.
	
	String name

	// Case 1: Creates wave uni
	//	Y1   Y2   Y3   Y4   Y5   Y6
	//	Y7   Y8   Y9   Y10  Y11  Y12
	if (isReal && isSingle && isEven)
		Wave Y1 = $StringFromList(0,rawWaveList)
		Wave Y2 = $StringFromList(1,rawWaveList) 
		Wave Y3 = $StringFromList(2,rawWaveList)
		Wave Y4 = $StringFromList(3,rawWaveList)
		Wave Y5 = $StringFromList(4,rawWaveList) 
		Wave Y6 = $StringFromList(5,rawWaveList)
		name = GenerateWaveName(header, sectionNumber, 0)		// Default: "uni<n>"
		Make /O /N=(numPointsInSection) $name
		Wave out = $name
		SetScale /P x, abcissaMin, abcissaIncrement, abcissaUnits, out
		SetScale d, 0, 0, ordinateUnits, out
		out[0,;6] = Y1[p/6]
		out[1,;6] = Y2[(p-1)/6]
		out[2,;6] = Y3[(p-2)/6]
		out[3,;6] = Y4[(p-3)/6]
		out[4,;6] = Y5[(p-4)/6]
		out[5,;6] = Y6[(p-5)/6]
		outputWaveList += name + ";"
	endif

	// Case 2: Creates waves uniX, uniY
	//	X1   Y1   X2   Y2   X3   Y3
	//	X4   Y4   X5   Y5   X6   Y6
	if (isReal && isSingle && isUnEven)
		Wave X1 = $StringFromList(0,rawWaveList)
		Wave X2 = $StringFromList(2,rawWaveList) 
		Wave X3 = $StringFromList(4,rawWaveList)
		name = GenerateWaveName(header, sectionNumber, 1)			// Default: "uni<n>X"
		Make /O /N=(numPointsInSection) $name
		Wave outX = $name
		SetScale d, 0, 0, abcissaUnits, outX
		outX[0,;3] = X1[p/3]
		outX[1,;3] = X2[(p-1)/3]
		outX[2,;3] = X3[(p-2)/3]
		outputWaveList += name + ";"

		Wave Y1 = $StringFromList(1,rawWaveList)
		Wave Y2 = $StringFromList(2,rawWaveList) 
		Wave Y3 = $StringFromList(5,rawWaveList)
		name = GenerateWaveName(header, sectionNumber, 2)			// Default: "uni<n>Y"
		Make /O /N=(numPointsInSection) $name
		Wave outY = $name
		SetScale d, 0, 0, ordinateUnits, outY
		outY[0,;3] = Y1[p/3]
		outY[1,;3] = Y2[(p-1)/3]
		outY[2,;3] = Y3[(p-2)/3]
		outputWaveList += name + ";"
	endif
	
	// Case 3:
	//	RY1  IY1  RY2  IY2  RY3  IY3
	//	RY4  IY4  RY5  IY5  RY6  IY6
	if (isComplex && isSingle && isEven)
		Wave RY1 = $StringFromList(0,rawWaveList)
		Wave RY2 = $StringFromList(2,rawWaveList)
		Wave RY3 = $StringFromList(4,rawWaveList) 
		name = GenerateWaveName(header, sectionNumber, 3)			// Default: "uni<n>Real"
		Make /O /N=(numPointsInSection) $name
		Wave outReal = $name
		SetScale /P x, abcissaMin, abcissaIncrement, abcissaUnits, outReal
		SetScale d, 0, 0, ordinateUnits, outReal
		outReal[0,;3] = RY1[p/3]
		outReal[1,;3] = RY2[(p-1)/3]
		outReal[2,;3] = RY3[(p-2)/3]
		outputWaveList += name + ";"

		Wave IY1 = $StringFromList(1,rawWaveList) 
		Wave IY2 = $StringFromList(3,rawWaveList)
		Wave IY3 = $StringFromList(5,rawWaveList)
		name = GenerateWaveName(header, sectionNumber, 4)			// Default: "uni<n>Imag"
		Make /O /N=(numPointsInSection) $name
		Wave outImag = $name
		SetScale /P x, abcissaMin, abcissaIncrement, abcissaUnits, outImag
		SetScale d, 0, 0, ordinateUnits, outImag
		outImag[0,;3] = IY1[p/3]
		outImag[1,;3] = IY2[(p-1)/3]
		outImag[2,;3] = IY3[(p-2)/3]
		outputWaveList += name + ";"
		
		if (createComplex)
			name = GenerateWaveName(header, sectionNumber, 5)		// Default: "uni<n>Complex"
			Make /O /C /N=(numPointsInSection) $name
			Wave/C outComplex = $name
			SetScale /P x, abcissaMin, abcissaIncrement, abcissaUnits, outComplex
			SetScale d, 0, 0, ordinateUnits, outComplex
			outComplex = cmplx(outReal, outImag)
			KillWaves outReal, outImag
			outputWaveList = name + ";"
		endif
	endif
	
	// Case 4:
	//	X1   RY1  IY1  X2  RY2  IY2
	//	X3   RY3  IY3  X4  RY4  IY4
	if (isComplex && isSingle && isUneven)
		Wave X1 = $StringFromList(0,rawWaveList)
		Wave X2 = $StringFromList(3,rawWaveList) 
		name = GenerateWaveName(header, sectionNumber, 1)		// Default: "uni<n>X"
		Make /O /N=(numPointsInSection) $name
		Wave outX = $name
		SetScale d, 0, 0, abcissaUnits, outX
		outX[0,;2] = X1[p/2]
		outX[1,;2] = X2[(p-1)/2]
		outputWaveList += name + ";"
		String xName1 = name

		Wave RY1 = $StringFromList(1,rawWaveList)
		Wave RY2 = $StringFromList(4,rawWaveList) 
		name = GenerateWaveName(header, sectionNumber, 3)		// Default: "uni<n>Real"
		Make /O /N=(numPointsInSection) $name
		Wave outReal = $name
		SetScale d, 0, 0, ordinateUnits, outReal
		outReal[0,;2] = RY1[p/2]
		outReal[1,;2] = RY2[(p-1)/2]
		outputWaveList += name + ";"

		Wave IY1 = $StringFromList(2,rawWaveList)
		Wave IY2 = $StringFromList(5,rawWaveList) 
		name = GenerateWaveName(header, sectionNumber, 4)		// Default: "uni<n>Imag"
		Make /O /N=(numPointsInSection) $name
		Wave outImag = $name
		SetScale d, 0, 0, ordinateUnits, outImag
		outImag[0,;2] = IY1[p/2]
		outImag[1,;2] = IY2[(p-1)/2]
		outputWaveList += name + ";"
		
		if (createComplex)
			name = GenerateWaveName(header, sectionNumber, 5)	// Default: "uni<n>Complex"
			Make /O /C /N=(numPointsInSection) $name
			Wave/C outComplex = $name
			SetScale d, 0, 0, ordinateUnits, outComplex 
			outComplex = cmplx(outReal, outImag)
			KillWaves outReal, outImag
			outputWaveList = xName1 + ";" + name + ";"
		endif
	endif
	
	// Case 5: Creates wave uni
	//	Y1     Y2     Y3     Y4
	//	Y5     Y6     Y7     Y8
	if (isReal && isDouble && isEven)	
		Wave Y1 = $StringFromList(0,rawWaveList)
		Wave Y2 = $StringFromList(1,rawWaveList) 
		Wave Y3 = $StringFromList(2,rawWaveList)
		Wave Y4 = $StringFromList(3,rawWaveList)
		Wave Y5 = $StringFromList(4,rawWaveList) 
		Wave Y6 = $StringFromList(5,rawWaveList)
		name = GenerateWaveName(header, sectionNumber, 0)		// Default: "uni<n>"
		Make /O /D /N=(numPointsInSection) $name
		Wave out = $name
		SetScale /P x, abcissaMin, abcissaIncrement, abcissaUnits, out
		SetScale d, 0, 0, ordinateUnits, out
		out[0,;6] = Y1[p/6]
		out[1,;6] = Y2[(p-1)/6]
		out[2,;6] = Y3[(p-2)/6]
		out[3,;6] = Y4[(p-3)/6]
		out[4,;6] = Y5[(p-4)/6]
		out[5,;6] = Y6[(p-5)/6]
		outputWaveList += name + ";"
	endif

	// Case 6: Creates waves uniX, uniY
	//	X1   Y1     X2   Y2
	//	X3   Y3     X4   Y4
	if (isReal && isDouble && isUnEven)
		Wave X1 = $StringFromList(0,rawWaveList)
		Wave X2 = $StringFromList(2,rawWaveList) 
		Wave X3 = $StringFromList(4,rawWaveList)
		name = GenerateWaveName(header, sectionNumber, 1)		// Default: "uni<n>X"
		Make /O /D /N=(numPointsInSection) $name
		Wave outX = $name
		SetScale d, 0, 0, abcissaUnits, outX
		outX[0,;3] = X1[p/3]
		outX[1,;3] = X2[(p-1)/3]
		outX[2,;3] = X3[(p-2)/3]
		outputWaveList += name + ";"

		Wave Y1 = $StringFromList(1,rawWaveList)
		Wave Y2 = $StringFromList(2,rawWaveList) 
		Wave Y3 = $StringFromList(5,rawWaveList)
		name = GenerateWaveName(header, sectionNumber, 2)		// Default: "uni<n>Y"
		Make /O /D /N=(numPointsInSection) $name
		Wave outY = $name
		SetScale d, 0, 0, ordinateUnits, outY
		outY[0,;3] = Y1[p/3]
		outY[1,;3] = Y2[(p-1)/3]
		outY[2,;3] = Y3[(p-2)/3]
		outputWaveList += name + ";"
	endif
	
	// Case 7:
	//	RY1    IY1    RY2    IY2
	//	RY3    IY3    RY4    IY4
	if (isComplex && isDouble && isEven)
		Wave RY1 = $StringFromList(0,rawWaveList)
		Wave RY2 = $StringFromList(2,rawWaveList)
		name = GenerateWaveName(header, sectionNumber, 3)		// Default: "uni<n>Real"
		Make /O /D /N=(numPointsInSection) $name
		Wave outReal = $name
		SetScale /P x, abcissaMin, abcissaIncrement, abcissaUnits, outReal
		SetScale d, 0, 0, ordinateUnits, outReal
		outReal[0,;2] = RY1[p/2]
		outReal[1,;2] = RY2[(p-1)/2]
		outputWaveList += name + ";"

		Wave IY1 = $StringFromList(1,rawWaveList) 
		Wave IY2 = $StringFromList(3,rawWaveList)
		name = GenerateWaveName(header, sectionNumber, 4)		// Default: "uni<n>Imag"
		Make /O /D /N=(numPointsInSection) $name
		Wave outImag = $name
		SetScale /P x, abcissaMin, abcissaIncrement, abcissaUnits, outImag
		SetScale d, 0, 0, ordinateUnits, outImag
		outImag[0,;2] = IY1[p/2]
		outImag[1,;2] = IY2[(p-1)/2]
		outputWaveList += name + ";"
		
		if (createComplex)
			name = GenerateWaveName(header, sectionNumber, 5)	// Default: "uni<n>Complex"
			Make /O /C /D /N=(numPointsInSection) $name
			Wave/C outComplex = $name
			SetScale /P x, abcissaMin, abcissaIncrement, abcissaUnits, outComplex
			SetScale d, 0, 0, ordinateUnits, outComplex
			outComplex = cmplx(outReal, outImag)
			KillWaves outReal, outImag
			outputWaveList = name + ";"
		endif
	endif
	
	// Case 8:
	//	X1   RY1    IY1
	//	X2   RY2    IY2
	if (isComplex && isDouble && isUneven)
		Wave X1 = $StringFromList(0,rawWaveList)
		Wave X2 = $StringFromList(3,rawWaveList) 
		name = GenerateWaveName(header, sectionNumber, 1)		// Default: "uni<n>X"
		Make /O /D /N=(numPointsInSection) $name
		Wave outX = $name
		SetScale d, 0, 0, abcissaUnits, outX
		outX[0,;2] = X1[p/2]
		outX[1,;2] = X2[(p-1)/2]
		outputWaveList += name + ";"
		String xName2 = name

		Wave RY1 = $StringFromList(1,rawWaveList)
		Wave RY2 = $StringFromList(4,rawWaveList) 
		name = GenerateWaveName(header, sectionNumber, 3)		// Default: "uni<n>Real"
		Make /O /N=(numPointsInSection) $name
		Wave outReal = $name
		SetScale d, 0, 0, ordinateUnits, outReal
		outReal[0,;2] = RY1[p/2]
		outReal[1,;2] = RY2[(p-1)/2]
		outputWaveList += name + ";"

		Wave IY1 = $StringFromList(2,rawWaveList)
		Wave IY2 = $StringFromList(5,rawWaveList) 
		name = GenerateWaveName(header, sectionNumber, 4)		// Default: "uni<n>Imag"
		Make /O /D /N=(numPointsInSection) $name
		Wave outImag = $name
		SetScale d, 0, 0, ordinateUnits, outImag
		outImag[0,;2] = IY1[p/2]
		outImag[1,;2] = IY2[(p-1)/2]
		outputWaveList += name + ";"
		
		if (createComplex)
			name = GenerateWaveName(header, sectionNumber, 5)	// Default: "uni<n>Complex"
			Make /O /C /D /N=(numPointsInSection) $name
			Wave/C outComplex = $name
			SetScale d, 0, 0, ordinateUnits, outComplex
			outComplex = cmplx(outReal, outImag)
			KillWaves outReal, outImag
			outputWaveList = xName2 + ";" + name + ";"
		endif
	endif

	return 0	
End

// LoadUniversalVersion58Section(pathName, fileName, sectionNumber, startLine, createComplex, quiet, GenerateWaveName, endLine)
// startLine is the zero-based line number where the section starts.
// Each section starts with a preamble - a line containing -1 and another line containing 58.
// It is followed by 11 headers lines and a variable number of data lines.
// This is followed by a -1 marking the end of the section.
// endLine is set to the line number of the line containing the -1 marking the end of the section.
static Function LoadUniversalVersion58Section(pathName, fileName, sectionNumber, startLine, createComplex, quiet, GenerateWaveName, endLine)
	String pathName		// Name of an Igor symbolic path or "".
	String fileName			// Name of file or full path to file.
	Variable sectionNumber	// Zero-based section number.
	Variable startLine		// Zero-based line on which the section starts.
	Variable createComplex	// 0: Create two real waves for complex data; 1: Create one complex wave for complex data
	Variable quiet			// 0: Print normal progress and error messages; 1: Print error messages only
	FUNCREF GenerateUniversal58WaveName GenerateWaveName	// Function for naming waves patterned after GenerateUniversal58WaveName
	Variable& endLine		// Output - zero-based line number where section ended
	
	endLine = -1

	String sectionNumberStr = num2istr(sectionNumber)
	
	Variable result
	String columnInfo

	// Load the first 13 lines which are header information
	String headerWaveName = "uni" + sectionNumberStr + "Header"
	
	columnInfo = "N=" + headerWaveName + ";"
	LoadWave /J /K=2 /L={0,startLine,13,0,0} /B=columnInfo /A /O /P=$pathName /Q fileName
	if (V_flag == 0)
		Printf "Failed to load header for section %d starting on line %d\r", sectionNumber, startLine
		return -1
	endif
	
	Wave/T uniHeader = $headerWaveName			// Get reference to text wave created by LoadWave
	struct UniversalFile58HeaderInfo header
	result = GetUniversal58HeaderInfo(sectionNumber, startLine, uniHeader, header)
	if (result != 0)
		return result
	endif
	// Print header		// For debugging only
	
	// Allow overriding name of header wave
	String newHeaderWaveName = GenerateWaveName(header, sectionNumber, 6) // Default: "uni<n>Header"
	if (CmpStr(newHeaderWaveName,headerWaveName) != 0)
		Duplicate /O $headerWaveName,  $newHeaderWaveName
		KillWaves /Z uniHeader
		Wave/T uniHeader = $newHeaderWaveName
	endif

	// Figure out how many lines of data to load
	
	Variable ordinateDataType			// 2=real SP, 4=real DP, 5=complex SP, 6=complex DP
	Variable numDataPairsOrValues		// Num XY pairs for uneven abcissa, num values for even abcissa
	Variable abcissaSpacing				// 0=uneven, 1=even (no abcissa values stored)

	// Determine characteristics of section data which depend on ordinate type (real or complex), the ordinate precision (single or double) and the abcissa spacing (even or uneven).
	Variable numDataColumns, numDataLines, defaultFieldWidth
	result = DetermineDataCharacteristics(header, numDataColumns, numDataLines, defaultFieldWidth, columnInfo)
	if (result != 0)
		return result
	endif
	
	// Section consists of 2-line preamble, 11 header lines, data lines, and -1 end-of-section marker
	endLine = startLine + 2 + 11 + numDataLines
	
	// Load the data.
	LoadWave /F={numDataColumns, defaultFieldWidth, 0} /K=0 /L={0,startLine+13,numDataLines,0,0} /A /O /B=columnInfo /P=$pathName /Q fileName
	if (V_flag == 0)
		Printf "Failed to load data for section %d starting on line %d\r", sectionNumber, startLine
		return -1
	endif
	String rawWaveList = S_waveNames		// Semicolon-separated list of loaded waves
	
	// This creates output waves from the raw loaded columns and sets wave properties
	String outputWaveList
	CreateOutputWaves(header, rawWaveList, sectionNumber, startLine, createComplex, GenerateWaveName, outputWaveList)
	
	if (!quiet)
		Printf "Loaded waves from section %d starting at line %d: %s\r", sectionNumber, startLine, outputWaveList
	endif
	
	// Kill raw waves
	Variable numItems = ItemsInList(rawWaveList)
	Variable i
	for(i=0; i<numItems; i+=1)
		String name = StringFromList(i, rawWaveList)
		KillWaves /Z $name
	endfor

	return 0		// Success
End

// CountLinesInFile(pathName, fileName)
// Returns the number of lines of text in the file.
// This counts includes the last line in the file which is empty.
// It is assumed that the only line in the file after the last end-of-section line
// is the final empty line which contains no characters.
static Function CountLinesInFile(pathName, fileName)
	String pathName, fileName
	
	Variable refNum
	Open /R /P=$pathName refNum as fileName
	
	String text
	Variable numLines = 0
	do
		FReadLine refNum, text
		if (strlen(text) == 0)			// Stop when we hit empty line or CR or LF or CRLF
			break
		endif
		numLines += 1
	while(1)
	
	return numLines
End

Function/S GenerateUniversal58DFName(filePath)
	String filePath					// Full path, partial path or file name
	
	String dfName
	
	String separator
	if (strsearch(filePath, ":", 0) > 0)
		separator = ":"		// Macintosh path
	else
		separator = "\\"	// Windows path
	endif
	String fileName = ParseFilePath(0, filePath, separator, 1, 0)	// Get file name by itself

	dfName = fileName
	Variable dotPos
	dotPos = strsearch(dfName, "." , 0)		// Remove extension if it is there.
	if (dotPos > 0)
		dfName = dfName[0, dotPos-1]
	endif
	dfName = "UF" + dfName				// Igor prefers names to start with a letter.
	dfName = CleanupName(dfName, 0)		// Make name kosher.

	return dfName
End

static Function CreateAndSetDataFolder(filePath, reuseDataFolder, GenerateDFName, dfName)
	String filePath					// Full path, partial path or file name
	Variable reuseDataFolder		// If true, any existing data is reused without prompting the user
	FUNCREF GenerateUniversal58DFName GenerateDFName
	String& dfName			// Output
	
	dfName = GenerateDFName(filePath)
	
	if (!reuseDataFolder && DataFolderExists(dfName))
		String promptStr
		sprintf promptStr, "Data folder '%s' exists. Reuse it anyway?", dfName
		DoAlert /T="Reuse Existing Data Folder?" 1, promptStr
		if (V_flag != 1)
			DoAlert /T="Data Folder Already Exists" 0, "Rename the conflicting data folder and try again."
			return -1			// User cancelled
		endif
	endif
	
	NewDataFolder /O /S $dfName
	
	return 0
End

// This structure is used when calling the LoadUniversalVersion58FilePI function (programmatic interface)
Structure UniversalFile58Params
	// Inputs
	String pathName				// Name of an Igor symbolic path or "".
	String fileName					// File name, partial path or full path to file.
	Variable reuseDataFolder		// If 1, existing data folder is reused without prompting. If 0, prompt in case of existing data folder.
	Variable createComplex			// 0: Create two real waves for complex data; 1: Create one complex wave for complex data.
	Variable quiet					// 0: Print normal progress and error messages; 1: Print error messages only.

	// Callback functions
	FUNCREF GenerateUniversal58DFName GenerateDFName			// Function to generate data folder name patterned after GenerateUniversal58DFName
	FUNCREF GenerateUniversal58WaveName GenerateWaveName	// Function to generate wave names patterned after GenerateUniversal58WaveName
	
	// Outputs
	DFREF dfr						// Set to point to the data folder created to hold the output waves
EndStructure

Function LoadUniversalVersion58FilePI(params)
	STRUCT UniversalFile58Params &params
	
	String pathName = params.pathName
	String fileName = params.fileName
	Variable reuseDataFolder = params.reuseDataFolder
	Variable createComplex = params.createComplex
	Variable quiet = params.quiet
	
	// First get a valid reference to a file.
	if ((strlen(pathName)==0) %| (strlen(fileName)==0))
		// Display dialog looking for file.
		String prompt = "Select version 58 universal file"
		Variable refNum
		Open /D /R /T=".unv????" /M=prompt /P=$pathName refNum as fileName
		fileName = S_fileName			// S_fileName is set by Open/D
		if (strlen(fileName) == 0)		// User cancelled?
			return -2
		endif
	endif

	Variable result
	
	String dfName
	FUNCREF GenerateUniversal58DFName GenerateDFName = params.GenerateDFName
	result = CreateAndSetDataFolder(fileName, reuseDataFolder, GenerateDFName, dfName)
	if (result != 0)
		return result
	endif
	params.dfr = :				// Output: Path to data folder created to hold waves
	
	// This is necessary so we can tell when we have loaded all of the sections.
	Variable numLinesInFile = CountLinesInFile(pathName, fileName)

	Variable startLine = 0
	Variable endLine			// Set by LoadUniversalVersion58Section
	Variable sectionNumber = 0
	
	do
		FUNCREF GenerateUniversal58WaveName GenerateWaveName = params.GenerateWaveName
		result = LoadUniversalVersion58Section(pathName, fileName, sectionNumber, startLine, createComplex, quiet, GenerateWaveName, endLine)
		if (result != 0)
			Printf "Error %d while loading section %d starting on line %d. Aborting load.\r", result, sectionNumber, startLine
			SetDataFolder ::		// Back to original data folder
			return result
		endif

		// break	// Load first section only during development
		
		startLine = endLine + 1
		if (startLine >= numLinesInFile-1)
			break			// All done
		endif
		
		sectionNumber += 1
	while(1)

	SetDataFolder ::		// Back to original data folder
	
	return 0
End

Function LoadUniversalVersion58File(pathName, fileName, reuseDataFolder, createComplex, quiet)
	String pathName			// Name of an Igor symbolic path or "".
	String fileName				// File name, partial path or full path to file.
	Variable reuseDataFolder	// If 1, existing data folder is reused without prompting. If 0, prompt in case of existing data folder.
	Variable createComplex		// 0: Create two real waves for complex data; 1: Create one complex wave for complex data.
	Variable quiet				// 0: Print normal progress and error messages; 1: Print error messages only.

	STRUCT UniversalFile58Params params
	params.pathName = pathName
	params.fileName = fileName
	params.reuseDataFolder = reuseDataFolder
	params.createComplex = createComplex
	params.quiet = quiet
	FUNCREF GenerateUniversal58DFName params.GenerateDFName = GenerateUniversal58DFName
	FUNCREF GenerateUniversal58WaveName params.GenerateWaveName = GenerateUniversal58WaveName

	Variable result = LoadUniversalVersion58FilePI(params)
	return result
End
